// ------------------------------------------------------------------------- //
// --- This example is a sequential version of the multiple-baseline     --- //
// --- stereo application                                                --- //
// --- Author: Anne Benoit                                               --- //
// --- Last modification: 01/10/2004                                     --- //
// ------------------------------------------------------------------------- //
#include <sys/times.h>
#include "../eSkel.c"
#include "stereo.h"

#define DEBUG  0     // set to 1 for debugging mode, 0 for normal mode
#define WINX   13    // window rows
#define WINY   13    // window columns

// For display
#define CASEDEAL 1
#define CASE3ST  2
#define CASESEQ  3
#define CASEMPI  4

double truethrput;
double st3secs[ITERS];  // times between completions at stage 3, 
                        // st3secs[0] must be ignored, leaving
                        // ITERS-1 interesting readings



// ------------------------------------------------------------------------- //
// getdata generate some input data. Pretty dull.
void getdata(int ref[ROWS+WINY][COLS], int m1[ROWS+WINY][COLS], int m2[ROWS+WINY][COLS],
	     float curbesterr[ROWS][COLS], int curbestdisp[ROWS][COLS]) {
  int i,j,ipos,jpos;
  int tardis,isize,jsize;

  //if (DEBUG) printf("Entering getdata\n");
  tardis = 12;
  isize=37;
  jsize=37;

  for (i=0;i<6;i++)
    for (j=0;j<COLS;j++) {
      ref[i][j]=0; 
      m1[i][j] =0; 
      m2[i][j] =0; 
    }
  
  for (i=6;i<ROWS+6;i++)
    for (j=0;j<COLS;j++) {
      ref[i][j]=i+j; 
      m1[i][j] =i+j; 
      m2[i][j] =i+j; 
    }
  
  for (i=ROWS+6;i<ROWS+13;i++)
    for (j=0;j<COLS;j++) {
      ref[i][j]=0; 
      m1[i][j] =0; 
      m2[i][j] =0; 
    }
  
  ipos = (ROWS-isize)/2 + 6 - 1;
  jpos = (COLS-jsize)/2 - tardis - 1;

  for (i=ipos;i<ipos+isize;i++)
    for (j=jpos;j<jpos+jsize;j++) {
      ref[i][j]         = i-ipos + j-jpos;
      m1[i][j+tardis]   = i-ipos + j-jpos;
      m2[i][j+2*tardis] = i-ipos + j-jpos;
    }
  
  for (i=0;i<ROWS;i++)
    for (j=0;j<COLS;j++) {
      curbesterr[i][j]  = 9999999.0;
      curbestdisp[i][j] = 0;
    }

}

// ------------------------------------------------------------------------- //
// gendiffimg forms the difference image
void gendiffimg(int ref[ROWS+WINY][COLS], int m1[ROWS+WINY][COLS], int m2[ROWS+WINY][COLS],
		float diffimg[ROWS+WINY][COLS], int curdisp) {
  int  i,j;

  for (i=0;i<ROWS+WINY;i++)
    for (j=0;j<COLS-2*curdisp;j++)
      diffimg[i][j] = (ref[i][j] - m1[i][j+curdisp])*(ref[i][j] - m1[i][j+curdisp]) +
	              (ref[i][j] - m2[i][j+2*curdisp])*(ref[i][j] - m2[i][j+2*curdisp]);
}

// ------------------------------------------------------------------------- //
// updatedispimg sums the difference image over a moving window and updates
// the current best error and best disparity images.
void updatedispimg(float diffimg[ROWS+WINY][COLS], float curbesterr[ROWS][COLS], 
		   int curbestdisp[ROWS][COLS], int curdisp) {
  int   i,j;
  float sum,csum[COLS];

  for (j=0;j<COLS;j++) {
    csum[j] = 0.0;
    for (i=0;i<WINX;i++)
      csum[j] = csum[j] + diffimg[i][j];
  }

  for (i=0;i<ROWS;i++) {
    sum = 0.0;
    for (j=0;j<7;j++)
      sum = sum + csum[j];
    
    if (sum < curbesterr[i][0]) {
      curbesterr[i][0]  = sum;
      curbestdisp[i][0] = curdisp;
    }

    for (j=1;j<7;j++) {
      sum = sum + csum[j+6];
      if (sum < curbesterr[i][j]) {
	curbesterr[i][j]  = sum;
	curbestdisp[i][j] = curdisp;
      }
    }

    for (j=7;j<COLS-6;j++) {
      sum = sum - csum[j-7];
      sum = sum + csum[j+6];
      if (sum < curbesterr[i][j]) {
	curbesterr[i][j]  = sum;
	curbestdisp[i][j] = curdisp;
      }
    }

    for (j=COLS-6;j<COLS;j++) {
      sum = sum - csum[j-7];
      if (sum < curbesterr[i][j]) {
	curbesterr[i][j]  = sum;
	curbestdisp[i][j] = curdisp;
      }
    }
    
    for (j=0;j<COLS;j++)
      csum[j] = csum[j] - diffimg[i][j] + diffimg[i+WINY][j];

  }
  

}

// ------------------------------------------------------------------------- //
// error displays an error message raised in testdata
void error(int i, int j, int tardis, int disp) {
  
  printf("*** ERROR in testdata at position (%d,%d). Expected disp %d but got %d *** \n",
	 i,j,tardis,disp);
  
}

// ------------------------------------------------------------------------- //
// testdata is testing output (matched with getdata)
void testdata(int disp[ROWS][COLS]) {

  int i,j,ipos,jpos;
  int tardis,isize,jsize;

  tardis = 12;
  isize=37;
  jsize=37;

  ipos = (ROWS-isize)/2 - 1 ;
  jpos = (COLS-jsize)/2 - tardis - 1;


  for (i=ipos+10;i<=ipos+isize-11;i++)
    for (j=jpos+10;j<=jpos+jsize-11;j++)
      if (disp[i][j]!=tardis) 
	error(i,j,tardis,disp[i][j]);
  
  for (i=0;i<=ipos-10;i++)
    for (j=0;j<=jpos+jsize+2*tardis+9;j++)
      if (disp[i][j]!=0) 
	error(i,j,0,disp[i][j]);
  
  for (i=0;i<=ipos+isize+9;i++)
    for (j=jpos+jsize+2*tardis+9;j<COLS;j++)
      if (disp[i][j]!=0) 
	error(i,j,0,disp[i][j]);
  
  for (i=ipos+isize+9;i<ROWS;i++)
    for (j=jpos-10;j<COLS;j++)
      if (disp[i][j]!=0) 
	error(i,j,0,disp[i][j]);

  for (i=ipos-10;i<ROWS;i++)
    for (j=0;j<=jpos-10;j++)
      if (disp[i][j]!=0) 
	error(i,j,0,disp[i][j]);

}





// ------------------------------------------------------------------------- //
void printperf(int runtype, int np, double dsecs, double esecs, int iter, double truethrput, double rawtime) {
  double secs;
  double dpercent, epercent;

  if (DEBUG)
    printf("Entering printperf, %f %f %d \n",dsecs,esecs,iter);

  dsecs = dsecs/iter;
  esecs = esecs/iter;
  secs  = dsecs+esecs;

  //printf("%f %f %f \n",dsecs,esecs,secs);
  
  dpercent = dsecs*100/secs;
  epercent = esecs*100/secs;

  //printf("%f %f \n",dpercent,epercent);

  printf("***********************************************\n");
  printf("* Multiple-Baseline Stereo (%3d x %3d)        *\n", ROWS, COLS);
  printf("***********************************************\n");
  printf("* Time for one image set:                     *\n");
  printf("* diff     : %9.6f (%2.0f percent)           *\n",dsecs,dpercent);
  printf("* error    : %9.6f (%2.0f percent)           *\n",esecs,epercent);
  printf("* total(s) : %9.6f                        *\n"    ,secs);
  printf("***********************************************\n\n");
  //printf("* Throughput :   %6.2f frames/sec            *\n",1.0/(dsecs+esecs));
  //printf("***********************************************\n");

  switch (runtype) {
    case CASEDEAL:
      printf("Deal%d,",np-2);
      break;
    case CASE3ST:
      printf("3ST,");
      break;
    case CASESEQ:
      printf("Seq,");
      break;
    case CASEMPI:
      printf("MPI,");
      break;
  }
  printf(" %d procs,", np);
  printf(" throughput was %4.2f frames/sec, run-time on P0 was %4.2f secs, %d images.\n",truethrput, rawtime, ITERS);
}




// ------------------------------------------------------------------------- //
int main (int argc, char *argv[])
{
  // input images
  int ref[ROWS+WINY][COLS];  // reference    image
  int  m1[ROWS+WINY][COLS];  // first  match image
  int  m2[ROWS+WINY][COLS];  // second match image

  // intermediate images
  float diffimg[ROWS+WINY][COLS];
  float curbesterr[ROWS][COLS];

  // output disparity image
  int curbestdisp[ROWS][COLS];

  // misc
  int i,j;
  int     count, curdisp;
  float   start;
  //float   dsecs, esecs, secs;
  //clock_t clk;
  double  mpitime, mpidsecs, mpiesecs, mpisecs, newsecs, oldsecs;
  double secs, runtime;
  int nprocs;

  // Initializing MPI and eSkel
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  SkelLibInit();
  //mpitime = MPI_Wtime();
  //printf("MPI time: %f\n",mpitime);

  // Starting the application
  printf("Entering the Multiple-Baseline Stereo application, Process %d\n", myrank());
  secs = MPI_Wtime();

  //dsecs = 0;
  mpidsecs = 0;
  //esecs = 0;
  mpiesecs = 0;
 
  oldsecs = 0;

  for (count=1; count<=ITERS; count++) {
    if (DEBUG) printf("Starting iteration %d\n", count);

    getdata(ref,m1,m2,curbesterr,curbestdisp);

    for (curdisp=0;curdisp<MAXDISP;curdisp++) {
      mpisecs = MPI_Wtime();
      gendiffimg(ref,m1,m2,diffimg,curdisp);
      mpidsecs = mpidsecs + MPI_Wtime() - mpisecs;


      mpisecs = MPI_Wtime();
      updatedispimg(diffimg,curbesterr,curbestdisp,curdisp);
      mpiesecs = mpiesecs + MPI_Wtime() - mpisecs;
    }

    testdata(curbestdisp);
    newsecs = MPI_Wtime();
    st3secs[count-1] = newsecs - oldsecs;
    oldsecs = newsecs;
  }
  
  truethrput = 0;
  for (i=1; i<ITERS; i++) truethrput += st3secs[i];
  truethrput = 1/(truethrput/(ITERS-1));
  
  // Print the total time for each process
  runtime = MPI_Wtime()-secs;

  printperf(CASESEQ,nprocs,mpidsecs,mpiesecs,ITERS,truethrput,runtime); 

  // Finalize MPI
  MPI_Finalize();
  return 0;
}
